/*
 * File:        uart.c
 * Description: 
 * Rev:         A
 * Created:     1997/12/11
 * $Author: kingdon $
 * mail:        kingdon@ptgrey.com
 *
 * Copyright Point Grey Research 1997
 * Vancouver B.C.
 *
 * The copyright to the computer program(s) herein
 * is the property of Point Grey Research Inc.
 * The program(s) may be used and/or copied only with
 * the written permission of Point Grey Research
 * or in accordance with the terms and conditions
 * stipulated in the agreement/contract under which
 * the program(s) have been supplied.
 *
 * $Id: uart.c,v 1.1 1997/12/16 22:03:20 kingdon Exp kingdon $
 */

#include <stdlib.h>
#include <rtems.h>
#include <bsp.h>
#include <rtems/libio.h>

#define INTERRUPT_EVENT	RTEMS_EVENT_1

struct UART_buf {
      volatile unsigned char* buffer;	/* buffer size must be a power of 2 */
      volatile unsigned int   ihead;	/* head of circular buffer */
      volatile unsigned int   itail;	/* tail of circular buffer */
      unsigned int            imask;	/* buflen - 1 */
      rtems_id                notify;	/* process to notify */
};

struct UART_info {
      unsigned	              init;	/* init state */
      long                    baud;
      int                     flow;
      struct UART_buf	      rcv;
      struct UART_buf         xmt;
      uartstruct*             dev;      /* device pointer */
};

#define ISFULL(B)	((unsigned)(((B).ihead-(B).itail))>(B).imask)
#define ISEMPTY(B)	((B).ihead==(B).itail)
#define UART_INIT(DATA)	{ DATA, 0, 0, sizeof(DATA)-1, 0 }

/* THESE SIZES MUST BE POWERS OF TWO!!! */
static unsigned char rcv0_data[1024];
static unsigned char xmt0_data[1024];
static unsigned char rcv1_data[1024];
static unsigned char xmt1_data[1024];

static struct UART_info UART_info[2] = {
   {
      NOINIT,
      0,
      0,
      UART_INIT(rcv0_data),
      UART_INIT(xmt0_data),
      &sim->uart1
   },
   {
      NOINIT,
      0,
      0,
      UART_INIT(rcv1_data),
      UART_INIT(xmt1_data),
      &sim->uart2
   }
};

#if 0
static struct UART_info* UART_info[2] = {
   &UART_info_table[0],
   &UART_info_table[1]
};
#endif

/*
 * Rcv interrupt...
 */
static rtems_isr
uart_intr( rtems_vector_number vec )
{
   unsigned int         chan = (vec - UART1_IRQ_VECTOR) & 1; /* this requires contiguous UART vectors */
   struct   UART_info*  info = &UART_info[chan];
   volatile uartstruct* uart = info->dev;

   /*
    * Notes:
    * - need to disable interrupts since a higher priority task might
    *   decide that it want to read...
    * - since interrupts are disabled UART_buf is no longer volatile and
    *   we can load ihead/itail into a register...
    * - this code should check for buffer overflows!
    */

#if 0
   fpga->led = 1;
#endif

   if( (uart->read.usr & 0x01) != 0 )
   {
      struct UART_buf*  rcv = &info->rcv;
      register unsigned ihead;
      register unsigned itail;
      register unsigned imask;
      register unsigned oldsr;

      _CPU_ISR_Disable( oldsr );
      ihead = rcv->ihead;
      itail = rcv->itail;
      imask = rcv->imask;

      while( (uart->read.usr & 0x01) != 0 ) {
	 rcv->buffer[ihead & imask] = uart->read.urbuf;
	 if ((ihead-itail<imask)) { /* HACK overflow protection -TDS */
	   ihead++;
	 }
	 else {
#if 0
	   RAW_FMT( "&");	   
	   _uart_flush(0);
#endif
	 }
      }

      rcv->ihead = ihead;
      _CPU_ISR_Enable( oldsr );

      /* potential race condition? */
      if( rcv->notify )
      {
	 rtems_id notify = rcv->notify;
	 rcv->notify = 0;
	 rtems_event_send( notify, INTERRUPT_EVENT );
      }
   }

   /*
    * Notes:
    * - while the xmt_buf is non-empty we can enable Tx interrupt
    * - same comments as for reading...
    */
   if( (uart->read.usr & 0x04) != 0 )
   {
      struct UART_buf*  xmt = &info->xmt;
      register unsigned itail;
      register unsigned ihead;
      register unsigned oldsr;

      _CPU_ISR_Disable( oldsr );
      ihead = xmt->ihead;
      itail = xmt->itail;
      while( ihead != itail && (uart->read.usr & 0x04) != 0 )
	 uart->write.utbuf = xmt->buffer[itail++ & xmt->imask];
      xmt->itail = itail;

      if( ihead == itail )
	 uart->write.uimr = 0x02;	/* disable TXRDY interrupt */
      else
	 uart->write.uimr = 0x03;	/* enable TXRDY interrupt */
      _CPU_ISR_Enable( oldsr );

      /* potential race condition? */
      if( xmt->notify )
      {
	 rtems_id notify = xmt->notify;
	 xmt->notify = 0;
	 rtems_event_send( notify, INTERRUPT_EVENT );
      }
   }

#if 0
   fpga->led = 0;
#endif

}

static int
uart_intr_enable( int chan, rtems_isr_entry uart_isr )
{
   volatile unsigned short* imr  = &sim->imr;
   rtems_isr_entry   old_isr;
   rtems_status_code status;

   if( chan > 1 )
	return 0;

   /* disable interrupts... */
   *imr |= (0x1000<<chan);

   if( uart_isr == 0 )
      return 1;

   status = rtems_interrupt_catch( uart_isr, UART1_IRQ_VECTOR+chan, &old_isr );

   /* enable interrupts... */
   *imr &= ~(0x1000<<chan);

   return status == RTEMS_SUCCESSFUL;
}

static inline void
uart_intr_RxReady( unsigned chan, int useRxReady )
{
   volatile uartstruct* uart = chan + sim->uart1;
   
   /*
    * set mode:
    *  umr1 = 0xF3 = RxRTS=1 RxIRQ=1 ERR=1 PM=10 PT=0 B/C=11
    *  umr1 = 0xB3 = RxRTS=1 RxIRQ=0 ERR=1 PM=10 PT=0 B/C=11
    * (see also _uart_init() )
    */
   useRxReady = 1;
   uart->write.ucr = 0x10;
   uart->write.umr = useRxReady ? 0xB3 : 0xF3;
}

/*
 * private interface to uart: used by console & rtems driver interface...
 */

void
_uart_flush( unsigned chan )
{
   register struct UART_info* info = &UART_info[chan%2];

   if( chan > 1 || info->init == NOINIT )
      return;

   while( !ISEMPTY(info->xmt) )
      uart_intr( UART1_IRQ_VECTOR+chan );
}

/*
 * There are potential race conditions here:
 *  - If a writing process decides to sleep because the xmt buffer is full
 *    but the interrupt handler manages to empty the buffer before
 *    the rtems_task_ident() call is made the process will not be woken
 *    up until the next time a uart interrupt occurs (which may be never)!
 *    This shouldn't be a problem if the xmt buffer if big enough.
 *  - If a reading process goes to sleep because the rcv buffer is empty
 *    but less than 4 characters arrive to the uart a rcv fifo full interrupt
 *    will not occur. This shouldn't be a problem if we set the Rx interrupt
 *    to be RxReady
 */
static int
await_notify( struct UART_buf* uart )
{
   rtems_event_set   pending;

   if( uart->notify != 0 )
      return 0;				/* too many waiting (impossible) */

   if( rtems_task_ident( RTEMS_SELF, 0, &uart->notify ) != RTEMS_SUCCESSFUL )
      return 0;				/* who am i? (impossible) */

   if( rtems_event_receive( INTERRUPT_EVENT,
			    RTEMS_WAIT|RTEMS_EVENT_ANY,
			    RTEMS_NO_TIMEOUT,
			    &pending ) != RTEMS_SUCCESSFUL )
      return 0;				/* await notify failed */

   return 1;
}

int
_uart_read( unsigned chan, char* buf, int maxlen, int term_chr )
{
   struct UART_info*         info = &UART_info[chan%2];
   register struct UART_buf* rcv = &info->rcv;
   register int              len = 0;

   if( chan > 1 ) {
     return -1;
   }

   if( info->init == NOINIT ) {
     _uart_init( chan, 0, RAW_INIT );
   }
      
   while( len < maxlen )
   {
      register unsigned itail;
      register unsigned oldsr;
      register int      chr;

      /* check if buf is empty */
      _CPU_ISR_Disable( oldsr );
      itail = rcv->itail;
      if( itail == rcv->ihead )
      {
	 /* yes: check if uart contains unread characters... */
	 _CPU_ISR_Enable( oldsr );
	 uart_intr( UART1_IRQ_VECTOR+chan );
	 /* recheck if buf is empty... */
	 _CPU_ISR_Disable( oldsr );
	 itail = rcv->itail;
	 uart_intr_RxReady( chan, 1 ); /* intr on any Rx */
	 if( itail == rcv->ihead )
	 {
	    /*
	     * yes: wait until uart is nonempty...
	     * NOTE: don't wait if within an interrupt handler!
	     */
	    _CPU_ISR_Enable( oldsr );
#if 0
	    if( chan == 1 ) RAW_FMT( "uart_rx(%d)", chan );
#endif
	    if( (oldsr & 0x0700) == 0 &&
		info->init == RTEMS_INIT &&
		!await_notify( rcv ) )
	    {
	       uart_intr_RxReady( chan, 0 ); /* intr only on RxFifo */
	       _uart_init( chan, 0, RAW_INIT );
	       break;
	    }
	    /* buffer is probably non-empty but lets check it again... */
	    uart_intr_RxReady( chan, 0 ); /* intr only on RxFifo */
	    continue;
	 }
	 uart_intr_RxReady( chan, 0 ); /* intr only on RxFifo */
      }

      /* This shouldnt happen (TSH) */
      if( ((rcv->ihead - itail) & rcv->imask ) == 0) {
	int _uart_write( unsigned chan, const char* buf, int maxlen );

	_CPU_ISR_Enable( oldsr );
#if 0
	RAW_FMT("[head=%x tail=%x]", rcv->ihead, itail);
	_uart_flush(0);
#elif 0
	_uart_write(0, "%", 1);
	_uart_flush(0);
#endif
      }

      /* finaly we've got at least one character... */
      else if( term_chr <= 0 )
      {
	 /*
	  * OPTIMIZATION:
	  * since we must read all maxlen chars lets advance
	  * itail, enable interrupts, and then copy them...
	  * THIS IS DANGEROUS because the buffer could overflow
	  * without our noticing it!
	  */
	 int npend = (rcv->ihead - itail) & rcv->imask;
	 if( npend > maxlen )
	    npend = maxlen;
	 rcv->itail = itail + npend;
	 _CPU_ISR_Enable( oldsr );
	 while( --npend >= 0 )
	    buf[len++] = rcv->buffer[itail++ & rcv->imask];
	 /*
	  * This hack is for Tyson: he wants the read to return with
	  * just whats available...
	  */
	 break;
      }
      else
      {
	 /* byte-by-byte case... */
	 chr = rcv->buffer[itail & rcv->imask];
	 rcv->itail = itail + 1;
	 _CPU_ISR_Enable( oldsr );
	 buf[len++] = chr;
	 if( chr < term_chr )
	    break;
      }
   }

   /* make sure the uart is awake! */
   uart_intr( UART1_IRQ_VECTOR+chan );
   return len;
}

int
_uart_write( unsigned chan, const char* buf, int maxlen )
{
   struct UART_info*         info = &UART_info[chan%2];
   register struct UART_buf* xmt = &info->xmt;
   register int              len = 0;

   if( chan > 1 )
      return -1;

   if( info->init == NOINIT ) {
     _uart_init( chan, 0, RAW_INIT );
   }

   while( len < maxlen )
   {
      register unsigned ihead;
      register unsigned oldsr;

      /* check if buf is full */
      _CPU_ISR_Disable( oldsr );
      ihead = xmt->ihead;
      if( (unsigned)(ihead-xmt->itail) > xmt->imask )
      {
	 /* yes: give check if uart contains unread characters... */
	 _CPU_ISR_Enable( oldsr );
	 uart_intr( UART1_IRQ_VECTOR+chan );
	 /* recheck if buf is full... */
	 _CPU_ISR_Disable( oldsr );
	 ihead = xmt->ihead;
	 if( (unsigned)(ihead-xmt->itail) > xmt->imask )
	 {
	    /*
	     * yes: wait until uart isn't full...
	     * NOTE: don't wait if within an interrupt handler!
	     */
	    _CPU_ISR_Enable( oldsr );
	    if( (oldsr & 0x0700) == 0 &&
		info->init == RTEMS_INIT &&
		!await_notify( xmt ) )
	    {
	       _uart_init( chan, 0, RAW_INIT );
	       break;
	    }
	    /* buffer probably isn't full but lets check it again... */
	    continue;
	 }
      }

      /*
       * finaly we've got some room...
       */
      xmt->buffer[ihead & xmt->imask] = buf[len++];
      xmt->ihead = ihead + 1;
      _CPU_ISR_Enable( oldsr );
   }

   /* make sure the uart is awake! */
   uart_intr( UART1_IRQ_VECTOR+chan );
   return len;
}

int
_uart_baud( unsigned chan, long baud )
{
   struct UART_info* info = &UART_info[chan%2];
   rtems_status_code status;

   if( chan > 1 )
      return 0;

   if( info->init != RTEMS_INIT )
      status = _uart_init( chan, baud, RAW_INIT );
   else
      status = _uart_init( chan, baud, RTEMS_INIT );

   return status == RTEMS_SUCCESSFUL;
}

rtems_status_code
_uart_init( unsigned _chan, long baud, unsigned level )
{
   unsigned                chan = _chan % 2;
   long                    clock_speed = _bootbsp_clock_speed();
   volatile uartstruct*    uart = (chan % 2) + &sim->uart1;
   volatile unsigned char* icr  = (chan % 2) + &sim->icr12;
   struct UART_info*       info = &UART_info[chan];
   unsigned                prev_level;
   long                    ubg;
   int                     i;

   if( _chan > 1 || level > RTEMS_INIT )
      return RTEMS_UNSATISFIED;

   if( baud <= 0 )
      baud = (info->baud <= 0 ) ? _bootbsp_baud_rate() : info->baud;

   /*
    * compute baud rate (using sysclock/32)...
    */
   if( chan == 0 && baud == 115200 )
      ubg = 0;			/* use slow pld buad clock... */
   else if( chan == 1 && baud == 230400 )
      ubg = 1;			/* use fast pld buad clock... */
   else
   {
      ubg = (clock_speed + 16 * baud) / (32 * baud);
      if( ubg < 2 )
	 return RTEMS_UNSATISFIED;
   }

   prev_level = info->init;

   if( level < prev_level )
   {
      /* cleanup request... */
      if( level < RTEMS_INIT && prev_level == RTEMS_INIT )
      {
	 info->init = RAW_INIT;
	 prev_level = RAW_INIT;
	 uart_intr_enable( chan, 0 );
      }
   }

   /* disallow uart interrupts during initialization... */
   uart_intr_enable( chan, 0 );

   if( level != NOINIT && (prev_level == NOINIT || baud != info->baud) )
   {
      /*
       * reset everything:
       *  reset cmds are misc[2:0] = 001b .. 101b
       */
      for( i = 5; i >= 1; i-- )
	 uart->write.ucr = (i<<4);

      /*
       * init interrupt:
       */
      if( chan == 0 )		
	 *icr = MCF5206_SIM_ICR_IL(UART1_IRQ_LEVEL)   |  /* AVEC=0 */
	        MCF5206_SIM_ICR_IP(UART1_IRQ_PRIORITY); 
      else
	 *icr = MCF5206_SIM_ICR_IL(UART2_IRQ_LEVEL)   |  /* AVEC=0 */
	        MCF5206_SIM_ICR_IP(UART2_IRQ_PRIORITY); 

      uart->write.uivr = UART1_IRQ_VECTOR+chan;	/* set intr vec */
      uart->write.uacr = 0x00;			/* disallow CTS interrupts */
      uart->write.uimr = 0x02;			/* enable FFULL interrupt */

      /*
       * setup baud rate:
       */
      if( ubg < 2 )
      {
	 *(char*)FAST_BAUD_REG = ubg;		/* set pld buad rate gen... */
	 uart->write.ucsr = 0xEE;		/* use tin0/16 */
      }
      else
      {
	 uart->write.ucsr = 0xDD;		/* use system_clock/32 */
	 uart->write.ubg1 = (ubg>>8) & 0xFF;	/* USB of baud rate */
	 uart->write.ubg2 = (ubg>>0) & 0xFF;	/* LSB of baud rate */
      }

      /*
       * set mode:
       *  umr1 = 0xF3 = RxRTS=1 RxIRQ=1 ERR=1 PM=10 PT=0 B/C=11
       *  umr2 = 0x17 = CM=00 TxRTS=0 TxCTS=1 SB=0111
       * (see also uart_intr_RxReady() )
       */
      uart->write.ucr = 0x10;			/* select umr1 */
      /* Full vs. Ready set here and other places */
      uart->write.umr = 0xB3;			/* 8 bit no parity */
      uart->write.umr = 0x17;			/* 1 stop bit */

      uart->write.ucr  = 0x05;			/* enable Tx & Rx */

      /*
       * resets don't seem to work: so input doesn't get flushed!
       */
      for( i = 0; i < 5 && (uart->read.usr & 0x01); i++ )
	 ubg = uart->read.urbuf;

      /*
       * CTS Hack: if CTS is not initially asserted clear the TxCTS bit...
       *    umr2 = 0x07 = CM=00 TxRTS=0 TxCTS=0 SB=0111
       */
      info->flow = (uart->read.uip & 0x01) == 0;
      info->flow = 0; /* Force no flow control */
      if( !info->flow ) {
	 uart->write.umr = 0x07;
      }

   }

   if( prev_level != NOINIT && level == NOINIT )
      level = RAW_INIT;		/* uart's are initialized! */

   info->init = level;
   info->baud = baud;

   /* only enable interrupts during rtems... */
   if( level == RTEMS_INIT )
      uart_intr_enable( chan, uart_intr );

   /* make sure the uart is awake! */
   uart_intr( UART1_IRQ_VECTOR+chan );
   return RTEMS_SUCCESSFUL;
}

#if 1
void
_uart_status( unsigned chan )
{
   struct UART_info  info;

   if( chan > 1 )
   {
      RAW_FMT( "uart%d: non-existant\n", chan );
      return;
   }

   info = UART_info[chan];
   RAW_FMT( "uart%d: init=%s baud=%d flow=%s read=%u(%u) write=%u(%u)\n",
	    chan,
	    (info.init == NOINIT ? "no" :
	     info.init == RAW_INIT ? "raw" : "rtems"),
	    info.baud, info.flow ? "on" : "off",
	    info.rcv.itail, info.rcv.ihead,
	    info.xmt.itail, info.xmt.ihead );
}
#endif

/*  uart_initialize
 *
 *  This routine initializes the console IO driver.
 *
 *  Input parameters: NONE
 *
 *  Output parameters:  NONE
 *
 *  Return values:
 */

rtems_device_driver uart_initialize(
   rtems_device_major_number  major,
   rtems_device_minor_number  minor,
   void                      *arg
)
{
   static const char *uart_name[] = {
      "/dev/uart0",
      "/dev/uart1"
   };
   static const char *console_name[] = {
      "/dev/uart",
   };

   rtems_status_code status;
   unsigned chan;
   unsigned i;

   for( chan = 0; chan < 2; chan++ )
   {
      status = _uart_init( chan, 0, RTEMS_INIT );
      if( status != RTEMS_SUCCESSFUL )
	 rtems_fatal_error_occurred( status );

      status = rtems_io_register_name( (char*)uart_name[chan],
				       major,
				       (rtems_device_minor_number)chan );
      if( status != RTEMS_SUCCESSFUL )
	 rtems_fatal_error_occurred( status );
   }

   chan = 0; /* hack the console chanel */
   for( i = 0; i < sizeof(console_name)/sizeof(*console_name); i++ )
   {
      status = rtems_io_register_name( (char*)console_name[i], major,
				       (rtems_device_minor_number)chan );
      if (status != RTEMS_SUCCESSFUL)
	 rtems_fatal_error_occurred(status);
   }

   return RTEMS_SUCCESSFUL;
}

/*
 *  Open entry point
 */
 
rtems_device_driver uart_open(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
   return (minor < 2) ? RTEMS_SUCCESSFUL : RTEMS_UNSATISFIED;
}
 
/*
 *  Close entry point
 */
 
rtems_device_driver uart_close(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
  return RTEMS_SUCCESSFUL;
}
 
/*
 * read bytes from a serial port.
 */

rtems_device_driver uart_read(
   rtems_device_major_number major,
   rtems_device_minor_number minor,
   void                    * arg
)
{
   rtems_libio_rw_args_t *rw_args = arg;
   int count;

   if( minor > 1 )
      return RTEMS_UNSATISFIED;

   count = _uart_read( (unsigned)minor,
		       (char*)rw_args->buffer,
		       (int)rw_args->count,
		       -1 );

  rw_args->bytes_moved = count;
  return (count >= 0) ? RTEMS_SUCCESSFUL : RTEMS_UNSATISFIED;
}
 
/*
 * write bytes to the serial port. Stdout and stderr are the same.
 */
 
rtems_device_driver uart_write(
   rtems_device_major_number major,
   rtems_device_minor_number minor,
   void                    * arg
)
{
   rtems_libio_rw_args_t *rw_args = arg;
   int count;

   if( minor > 1 )
      return RTEMS_UNSATISFIED;

   count = _uart_write( (unsigned)minor,
			(char*)rw_args->buffer,
			(int)rw_args->count );

  rw_args->bytes_moved = count;
  return (count >= 0) ? RTEMS_SUCCESSFUL : RTEMS_UNSATISFIED;
}

/*
 *  IO Control entry point
 */
 
rtems_device_driver uart_control(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
  return RTEMS_SUCCESSFUL;
}

